﻿/*
event
	create
	stay
	move
	
	integrate
	disappear
*/

import flash.display.BitmapData;
import flash.geom.Matrix;
import flash.geom.Rectangle;
import flash.geom.ColorTransform;
import flash.filters.ColorMatrixFilter;
import com.quasimondo.geom.ColorMatrix;
import flash.filters.ConvolutionFilter;
import mx.events.EventDispatcher;

class com.kampei.image.MotionDetector
{
	private var original_width:Number;
	private var original_height:Number;
	private var resolution_scale:Number;
	private var _min_size_new_rect_rate:Number;
	private var block_min_size:Number;
	private var _noize_min_pix:Number;
	private var _rect_life:Number;
	private var check_interval:Number;
	private var _threshold:Number;
	private var _rect_inflation:Number;
	
	private var next_bmp:BitmapData;
	private var now_bmp:BitmapData;
	private var difTest_bmp:BitmapData;

	private var targetImage; //MovieClip or BitmapData
	private var noiseReduction:ConvolutionFilter;
	
	private var gray:ColorMatrix;
	private var grayCm:ColorMatrixFilter;
	private var scaleMatrix:Matrix;
	private var bound:Rectangle;
	
	private var colorDiffs:Array;
	private var idCounter:Number;
	private var recs:Array;
	private var recs_watching:Array;
	private var recs_preborn:Array;
	
	private function dispatchEvent(e:Object){};
	public function addEventListener(event:String, listener:Object){};
	
	private var checkIntervalID:Number

	private var rectDump0:Array;
	private var rectDump1:Array;
	private var dumpDiff:Function = null;
	private var dumpLabel:Function = null;

	//画面の大きさに対する、新規rectを動体と認める最小限の大きさの割合(ノイズ的な動体を排除するため)
	public function get min_size_new_rect_rate():Number{ return _min_size_new_rect_rate; };
	public function set min_size_new_rect_rate(val:Number){ _min_size_new_rect_rate = val; };
	
	//動体が検知されなくなってから、その動体を消去するまでのステップ数
	public function get rect_life():Number{ return _rect_life; };
	public function set rect_life(val:Number){ _rect_life = val; };
	
	//差分画面を白黒2値化する時の閾値
	public function get threshold():Number{ return _threshold; };
	public function set threshold(val:Number){ _threshold = val; };
	
	//画面からノイズを除去する際のピクセルの最小限の大きさ
	public function get noize_min_pix():Number{ return _noize_min_pix; };
	public function set noize_min_pix(val:Number){ noiseReduction.bias = -0x100 * val; };
	
	//動体部分を矩形として判断するさいの、個々の矩形の大きさ
	public function get rect_inflation():Number{ return _rect_inflation; };
	public function set rect_inflation(val:Number){ _rect_inflation =  val; };
	
	/*
	w,h: 検知元画像の大きさ
	scale: 検知元画像をどれくらいの大きさでチェックするか。等倍だと精度が高いと思うけど、処理負荷も高い
	interval: 何ミリセックごとにチェックするか
	*/
	function MotionDetector(w:Number, h:Number, scale:Number, interval:Number){
		
		EventDispatcher.initialize(this);
		
		original_width = w;
		original_height = h;
		_threshold = 0xff101010;
		resolution_scale = scale;
		check_interval = interval;
		_min_size_new_rect_rate = 0.02;
		_noize_min_pix = 7;
		_rect_life = 5;
		_rect_inflation = 4;
		
		next_bmp = new BitmapData(original_width * resolution_scale, original_height * resolution_scale, false);
		now_bmp = new BitmapData(original_width * resolution_scale, original_height * resolution_scale, false);
		difTest_bmp = new BitmapData(original_width, original_height, false);
		
		noiseReduction = new ConvolutionFilter();
		noiseReduction.matrixX = 3;
		noiseReduction.matrixY = 3;
		noiseReduction.matrix = [
			1, 1, 1,
			1, 0, 1,
			1, 1, 1
		];
		noiseReduction.bias = -0x100 * _noize_min_pix;
		
		gray = new ColorMatrix();
		gray.adjustSaturation(0);
		grayCm = new ColorMatrixFilter(gray.matrix);
		
		scaleMatrix = new Matrix();
		scaleMatrix.scale(resolution_scale, resolution_scale);
		
		bound = new Rectangle(0, 0, original_width*resolution_scale, original_height*resolution_scale);
		
		colorDiffs = [];
		idCounter = 0;
		recs = [];
		recs_watching = [];
		recs_preborn = [];
	}
	
	public function startDetect(img){
		//trace("startDetect: " + img);
		targetImage = img;
		checkIntervalID = setInterval(this, "check", check_interval);
	}
	
	public function stopDetect(){
		//trace("stopDetect");
		clearInterval(checkIntervalID);
	}
	
	private function check(){
		//trace("check");
		recs = [];
		var recs_candidates = [];
		var recs_disappeared = [];
		for(var i in recs_watching){
			recs_watching[i].isects = [];
		}
		
		// ************** BitmapDataの処理 **************
		
		//checkVideoUpdated
		difTest_bmp.draw(targetImage);
		var cnt:Number= 0;
		var changed:Boolean = false;
		for(var h=1; h<4; h++){
			for (var v=1; v<3; v++){
				var colorNow:Number = difTest_bmp.getPixel(original_width/4*h, original_height/3*v);
				if(colorNow != colorDiffs[cnt]) changed = true;
				colorDiffs[cnt] = colorNow
				cnt ++;
			}
		}
		if(!changed)	return;
		
		//差分検出
		now_bmp.draw(next_bmp);
		now_bmp.draw(targetImage, scaleMatrix, new ColorTransform(), 7);	
		next_bmp.draw(targetImage, scaleMatrix);
		
		//グレイスケール
		now_bmp.applyFilter(now_bmp, bound, null, grayCm);

		//2階調化
		now_bmp.threshold(now_bmp, bound, null, "<=", _threshold, 0xff000000);
		now_bmp.threshold(now_bmp, bound, null, "!=", 0xFF000000, 0xFFFFFFFF);

		//ノイズリダクション
		var bmp_noise:BitmapData = now_bmp.clone();
		bmp_noise.applyFilter(bmp_noise, bound, null, noiseReduction);
		now_bmp.threshold(bmp_noise, bound, null, "==", 0xff000000, 0xff000000, 0xFF);

		//dump
		dumpDiff(now_bmp);
		
		//ブロック化
		while(true){
			var wbr = now_bmp.getColorBoundsRect(0xFFFFFF, 0xFFFFFF);
			if(wbr.isEmpty()) break;
			var x = wbr.x;
			for (var y = wbr.y + wbr.height - 1; y >= wbr.y; --y) {
				if (now_bmp.getPixel(x, y) == 0xFFFFFF) {
					now_bmp.floodFill(x, y, 0xFF00FF);
					var br = now_bmp.getColorBoundsRect(0xFFFFFF, 0xFF00FF);
					br.inflate(_rect_inflation,_rect_inflation);
					now_bmp.fillRect(br, 0x0000ff);
					recs.push(br);
					break;
				}
			}
		}
		
		//dump
		dumpLabel(now_bmp);
		
		rectDump0 = recs.slice();
		
		
		// ************** rectangleの処理 **************
		
		//rectangleが重なってたら、同じものとみなして、数を絞る
		for(var idx=0; idx<recs.length; idx++){
			var idx_rec = recs[idx];
			for(var i=idx+1; i<recs.length; i++){
				var tgt_rec = recs[i];
				if( idx_rec!=tgt_rec && idx_rec.intersects(tgt_rec) ){
					recs.push(idx_rec.union(tgt_rec));
					recs.splice(Number(i), 1);
					recs.splice(Number(idx), 1);
					idx --;
					break;
				}
			}
		}
		
		rectDump1 = recs.slice();
		
		//今回検出分とwatchしてる分の重なりを検証
		for(var i=0; i<recs.length; i++){
			var rec_ref = recs[i];
			if(rec_ref.width*rec_ref.height >= original_height*original_width*resolution_scale*resolution_scale*0.7){
				//画面全体が変化している場合(カメラの自動露出調整、オートホワイトバランス調整など)は無視。
				recs.splice(Number(i), 1);
				i--;
			}else{
				rec_ref.isects = [];
				for(var j in recs_watching){
					var rec_watch = recs_watching[j];
					if(rec_watch.intersects(rec_ref)){
						rec_watch.isects.push(rec_ref);
						rec_ref.isects.push(rec_watch);
					}
				}
			}
		}
		
		//今回検出分がそれぞれ、いくつのwatch分に重なってるか
		for(var i=0; i<recs.length; i++){
			var rec_ref = recs[i];
			var isectsNum = rec_ref.isects.length;
			
			//1つもwatch中のrectとintersectしてない → 新規追加
			if(isectsNum == 0){
				if(rec_ref.width*rec_ref.height <= original_width*original_height*resolution_scale*_min_size_new_rect_rate){
					//結構小さい → 廃棄
					//trace("tooSmall"+(rec_ref.width*rec_ref.height)+":"+(original_width*original_height*resolution_scale*_min_size_new_rect_rate));
					recs.splice(Number(i), 1);
					i--;
				}else{
					//新規生成
					rec_ref.id = ++idCounter;
					rec_ref.event = "create";
					rec_ref.life = _rect_life;
					recs_candidates.push(rec_ref);
				}
			
			//複数のwatch中のrectとintersectしている → watch中のrectと最もたくさんintersectしているものに統合
			}else if(isectsNum >= 2){
				var rect_integrate = null;
				for(var j=0; j<isectsNum-1; j++){
					var rec_now = rec_ref.isects[j];
					var rec_next = rec_ref.isects[j+1];
					rect_integrate = (rec_now.isects.length >= rec_next.isects.length) ? rec_now : rec_next;
				}
				for(var j=0; j<isectsNum; j++){
					if(rect_integrate != rec_ref.isects[j]){
						var rec_disappear = rec_ref.isects[j];
						rec_disappear.event = "integrate";
						rec_disappear.integarateTarget =  rect_integrate;
						recs_disappeared.push(rec_disappear);
						//recs_watchingから統合されたrectを削除
						for(var k=0; k<recs_watching.length; k++){
							if(recs_watching[k] == rec_disappear){
								recs_watching.splice(k, 1);
								break;
							}
						}
					}
				}
			}
		}
		
		//重なりデータを下に、新規watch候補を選定
		for(var i=0; i<recs_watching.length; i++){
			var rec_watch = recs_watching[i];
			var isectsNum = rec_watch.isects.length;
			if(isectsNum == 0){
				//0個の場合 → LIFE分継続
				//trace("watch_未intersect: " + rec_watch.id + " life:"+ rec_watch.life + " event: " + rec_watch.event);
				rec_watch.life -- ;
				rec_watch.event = "stay";
				recs_candidates.push(rec_watch);
			}else if(isectsNum == 1){
				//一つの場合 → 維持
				//trace("watch_１つintersect: " + rec_watch.id  + " life:" + rec_watch.life);
				rec_watch.isects[0].id = rec_watch.id;
				rec_watch.isects[0].event = "move";
				rec_watch.isects[0].life = _rect_life;
				recs_candidates.push(rec_watch.isects[0]);
				
			}else{
				//複数個の場合 → union
				//それぞれの距離を測ったほうがいいかも
				//あるいは、一つのrectの面積の上限を決めるとか
				//(2つのrectが一旦unionされると以降つながり続ける→すれ違った人が、ずっと同じ人間だと認識されてしまうだろう)
				//計算面倒だなあ
				//trace("watch_複数にintersect: id:" + rec_watch.id + " 個数:" + rec_watch.isects.length + " life:" + rec_watch.life);
				while(rec_watch.isects.length > 1){
					rec_watch.isects[0] = rec_watch.isects[0].union(rec_watch.isects[1]);
					rec_watch.isects.splice(1, 1);
				}
				rec_watch.isects[0].id = rec_watch.id;
				rec_watch.isects[0].event = "move";
				rec_watch.isects[0].life = _rect_life;
				recs_candidates.push(rec_watch.isects[0]);
			}
		}
		recs_watching = [];
		
		//候補から、判定
		for(var i=0; i<recs_candidates.length; i++){
			var rec_item = recs_candidates[i];
			
			var infratex = (rec_item.width - _rect_inflation <= _rect_inflation) ? 0 : _rect_inflation;
			var infratey = (rec_item.height - _rect_inflation <= _rect_inflation) ? 0 : _rect_inflation;
			rec_item.inflate(-infratex, -infratey);
			
			//誰ともintersectしてない寿命切れのrectを削除
			if(rec_item.event == "stay"){
				if(rec_item.life <= 0){
					//trace("disappear : " + rec_item.id);
					rec_item.event = "disappear";
					recs_disappeared.push(rec_item);
				}else{
					recs_watching.push(rec_item);
				}
			//移動、統合物
			}else if(rec_item.event == "move"){
				recs_watching.push(rec_item);
			
			//新規
			}else if(rec_item.event == "create"){
				recs_watching.push(rec_item);
			}
		}

		// dispatch events
		for(var i=0; i<recs_watching.length; i++){
			var tgt_rect = recs_watching[i];
			dispatch(tgt_rect.event, tgt_rect.id, tgt_rect);
		}
		for(var i=0; i<recs_disappeared.length; i++){
			var tgt_rect = recs_disappeared[i];
			dispatch(tgt_rect.event, tgt_rect.id, tgt_rect);
		}
		dispatchEvent({type:"update"});
	}
	
	private function dispatch(event:String, id, rect){
		var forResize = rect.clone();
		forResize.x = Math.round(forResize.x / resolution_scale);
		forResize.y = Math.round(forResize.y / resolution_scale);
		forResize.width = Math.round(forResize.width / resolution_scale);
		forResize.height = Math.round(forResize.height / resolution_scale);
		dispatchEvent({type:event, id:id, target:forResize}); //integrate,disappear
	}
}



